import shutil
from pathlib import Path

from autograder.config_manager import ALLOWED_LANGUAGES

from .autograder import AutograderPaths
from .testcases import ALLOWED_LANGUAGES


def create_dir(path: Path):
    if not path.exists():
        print(f"{path.name} directory does not exist. Creating...")
        path.mkdir()
    else:
        print(f"Found {path.name} directory")


def main(paths: AutograderPaths):
    if not paths.current_dir.exists() or not paths.current_dir.is_dir():
        print(f"Directory {paths.current_dir} not found. Please, use an existing directory.")
        exit(0)
    print("Hello. I will now guide you through the initial setup of autograder.")
    ans = input(f"Would you like to grade submissions located in '{paths.current_dir}'? (Yes/No) ")
    if not ans.lower().startswith("y"):
        print("You probably haven't specified a directory to the grader. Use 'autograder path/to/submissions/dir'")
        exit(0)
    create_dir(paths.tests_dir)
    create_dir(paths.testcases_dir)
    create_dir(paths.input_dir)
    create_dir(paths.output_dir)
    create_dir(paths.extra_dir)
    config_path = paths.config
    if not config_path.exists():
        print(f"config.ini not found in {paths.tests_dir}. Creating a default config...")
        paths.generate_config()
    else:
        print("Found config.ini")
    stdout_formatters_path = paths.stdout_formatters
    if not stdout_formatters_path.exists():
        print(f"{stdout_formatters_path.name} not found. Creating a default file...")
        shutil.copy(paths.default_stdout_formatters, str(paths.stdout_formatters))
    else:
        print("Found stdout_formatters.py")

    ans = input(
        "You are now ready to start working with autograder.\n"
        "Would you like me to generate the testcase templates? (Yes/No) "
    )
    if ans.lower().startswith("y"):
        allowed_languages = ", ".join(ALLOWED_LANGUAGES.keys())
        while True:
            choice = input(
                f"Choose a programming language you'd like to generate testcase templates for ({allowed_languages}): "
            )
            lang = ALLOWED_LANGUAGES.get(choice, None)
            if lang is None:
                print(f"Couldn't find the language with name '{choice}'. Please, try again.")
            else:
                break
        shutil.copytree(Path(__file__).parent / "templates" / choice, paths.testcases_dir, dirs_exist_ok=True)
    print(
        "\n\nNow if you want to grade your submissions, you can use 'autograder path/to/submissions/dir' "
        "for this directory."
    )
    print(f"You can write your testcases in {paths.testcases_dir}")
    print("If you want to see command line options, use 'autograder -h'")
    print(f"You can put the stdin inputs to your testcases into {paths.input_dir}")
    print(f"You can put the expected outputs to your testcases into {paths.output_dir}")
    print(f"You can put the extra files to be available for each testcase into {paths.extra_dir}")
    print(f"You can configure grading by editing {paths.config}")
    print("You can find readme at https://github.com/Ovsyanka83/autograder")



from .testcases.util.testcase_io import TestCaseIO
from .testcases.multifile import MultifileTestCase, is_multifile_submission
import multiprocessing
import os
import shutil
from contextlib import contextmanager
from pathlib import Path
from stat import S_IRGRP, S_IROTH, S_IRUSR, S_IWUSR, S_IXUSR
from typing import Callable, Dict, List, Set, Tuple, Type

import sh

from .testcases.abstract_base_class import ArgList, TestCase, submission_is_allowed

from .testcases import Submission
from .config_manager import GradingConfig
from .output_summary import BufferOutputLogger, GradingOutputLogger, get_submission_name
from .util import AutograderError, import_from_path

READ_EXECUTE_PERMISSION = S_IRUSR ^ S_IRGRP ^ S_IROTH ^ S_IXUSR
READ_EXECUTE_WRITE_PERMISSION = READ_EXECUTE_PERMISSION ^ S_IWUSR


class Grader:
    no_output: bool
    submissions: List[Submission]
    tests: Dict[Type[TestCase], List[TestCase]]
    raw_submissions: List[Path]
    paths: "AutograderPaths"
    multifile_submissions_found: bool

    def __init__(self, current_dir, no_output=False, submissions=None):
        if submissions is None:
            submissions = []
        self.no_output = no_output
        self.raw_submissions = submissions
        self.stdout_formatters = {}
        self.paths = AutograderPaths(current_dir)

    def run(self):
        io_choices = {}
        try:
            self.stdout_formatters = self._import_formatters(self.paths.stdout_formatters)
            self.config = GradingConfig(self.paths.testcases_dir, self.paths.config, self.paths.default_config)
            self.logger = GradingOutputLogger(
                self.paths.current_dir,
                self.paths.output_summary,
                self.paths.results_dir,
                self.config.assignment_name,
                self.config.total_points_possible,
                self.no_output,
                self.config.generate_results,
            )
            self._prepare_directory_structure()
            self.submissions, self.multifile_submissions_found = self._gather_submissions(self.raw_submissions)
            io_choices = self._gather_io()
            self.tests = self._gather_testcases(io_choices)  # also copies testcases to temp dir
            if self.multifile_submissions_found:
                self.tests[MultifileTestCase] = self._generate_multifile_testcases(io_choices)
            process_count = multiprocessing.cpu_count() if self.config.parallel_grading_enabled else 1
            with multiprocessing.Pool(process_count) as pool:
                man = multiprocessing.Manager()
                total_class_points = sum(pool.map(Runner(self, man.Lock()), self.submissions))
            class_average = total_class_points / len(self.submissions)
            self.logger(f"\nAverage score: {round(class_average)}/{self.config.total_points_possible}")
            self.logger.print_key()
        finally:
            self.cleanup()
            for io in io_choices.values():
                io.cleanup()
        return class_average

    def run_on_single_submission(self, submission: Submission, lock):
        with temporarily_change_dir(submission.dir):
            self._copy_extra_files(submission.dir)
            with self.logger.single_submission_output_logger(lock) as logger:
                result = self._get_testcase_output(submission, logger)
        # Cleanup after running tests on student submission
        shutil.rmtree(submission.dir)
        return result

    def cleanup(self):
        if self.paths.temp_dir.exists():
            shutil.rmtree(self.paths.temp_dir)

    def _prepare_directory_structure(self):
        # Cleanup in case any error
        self.cleanup()
        self.paths.temp_dir.mkdir()
        self._check_required_directories_exist()
        if self.config.generate_results:
            self.paths.results_dir.mkdir(exist_ok=True)

    def _copy_extra_files(self, to_dir: Path):
        if self.paths.extra_dir.exists():
            for path in self.paths.extra_dir.iterdir():
                new_path = to_dir / path.name
                shutil.copy(str(path), str(new_path))

    def _check_required_directories_exist(self):
        for directory in self.paths.required_dirs:
            if not directory.exists():
                raise AutograderError(
                    f"{directory} directory not found. It is required for the grader to function.\n"
                    "Maybe you specified the incorrect directory? Use `autograder submission_directory_here`"
                )

    def _gather_submissions(self, submissions_to_grade) -> Tuple[List[Submission], bool]:
        """ Returns sorted list of paths to submissions """
        submissions_to_grade = set(submissions_to_grade)
        submissions: List[Submission] = []
        multifile_submissions_found = False
        for submission_path in self.paths.current_dir.iterdir():
            if submissions_to_grade and submission_path.name not in submissions_to_grade:
                continue

            testcase_type = self.config.testcase_types.get(submission_path.suffix, None)
            if testcase_type is not None:
                if self.config.auto_source_file_name_enabled:
                    submissions.append(Submission(submission_path, testcase_type, self.paths.temp_dir))
                elif submission_is_allowed(
                    submission_path,
                    self.config.possible_source_file_stems,
                    self.config.source_file_stem_is_case_insensitive,
                ):
                    submissions.append(Submission(submission_path, testcase_type, self.paths.temp_dir))
                else:
                    self.logger(f"{submission_path} does not contain the required file name. Skipping it.")
            elif (
                self.config.multifile_submissions_enabled
                and submission_path.is_dir()
                and is_multifile_submission(
                    submission_path,
                    self.config.possible_source_file_stems,
                    self.config.source_file_stem_is_case_insensitive,
                )
            ):
                submissions.append(Submission(submission_path, MultifileTestCase, self.paths.temp_dir))
                multifile_submissions_found = True

        if not len(submissions):
            raise AutograderError(f"No student submissions found in '{self.paths.current_dir}'.")

        # Allows consistent output
        submissions.sort(key=lambda s: s.path)
        return submissions, multifile_submissions_found

    def _gather_io(self) -> Dict[str, TestCaseIO]:
        GET_FILE_STEMS = lambda d: (p.stem for p in d.iterdir()) if d.exists() else []
        outputs = GET_FILE_STEMS(self.paths.output_dir)
        inputs = GET_FILE_STEMS(self.paths.input_dir)
        io: Set[str] = set().union(outputs, inputs)
        dict_io = {p: TestCaseIO(p, self.stdout_formatters, self.paths.input_dir, self.paths.output_dir) for p in io}
        return dict_io

    def _generate_multifile_testcases(self, io_choices: Dict[str, TestCaseIO]) -> List[TestCase]:
        default_timeout = self.config.timeouts.get("ALL", 1)
        default_weight = self.config.testcase_weights.get("ALL", 1)
        return [
            MultifileTestCase(
                self.paths.temp_dir,
                self.config.timeouts.get(io.name, default_timeout),
                {},
                self.config.anti_cheat,
                self.config.testcase_weights.get(io.name, default_weight),
                io_choices,
            )
            for io in io_choices.values()
            if io.expected_output
        ]

    def _gather_testcases(self, io) -> Dict[Type[TestCase], List[TestCase]]:
        """ Returns sorted list of testcases from tests/testcases """
        tests = {t: [] for t in self.config.testcase_types.values()}
        default_weight = self.config.testcase_weights.get("ALL", 1)
        default_timeout = self.config.timeouts.get("ALL", 1)
        for test in self.paths.testcases_dir.iterdir():
            weight = self.config.testcase_weights.get(test.name, default_weight)
            timeout = self.config.timeouts.get(test.name, default_timeout)
            testcase_type = self.config.testcase_types.get(test.suffix, None)
            if not test.is_file():
                continue
            if testcase_type is None:
                self.logger(f"No appropriate language for {test} found.")
                continue
            arglist = self.config.generate_arglists(test.name)
            shutil.copy(test, self.paths.temp_dir)
            tests[testcase_type].append(
                testcase_type(  # type: ignore # The typing error here appears due to the limitations of python's typing
                    self.paths.temp_dir / test.name,
                    timeout,
                    arglist,
                    self.config.anti_cheat,
                    weight,
                    io,
                )
            )
        # Allows consistent output
        for test_list in tests.values():
            test_list.sort(key=lambda t: t.path.name)
        return tests

    def _import_formatters(self, path_to_stdout_formatters: Path) -> Dict[str, Callable[[str], str]]:
        if path_to_stdout_formatters.exists():
            module = import_from_path("stdout_formatters", path_to_stdout_formatters)
            return {k: v for k, v in module.__dict__.items() if callable(v)}
        else:
            return {}

    def _precompile_submission(self, submission: Submission, logger):
        try:
            arglists = self.config.generate_arglists(submission.path.name)
            return submission.type.precompile_submission(
                submission.path,
                submission.dir,
                self.config.possible_source_file_stems,
                self.config.source_file_stem_is_case_insensitive,
                arglists[ArgList.SUBMISSION_PRECOMPILATION],
            )
        except Exception as e:  # type: ignore
            self.logger.print_precompilation_error_to_results_file(
                submission, self.config.total_points_possible, e, logger
            )

    def _get_testcase_output(self, submission: Submission, logger: BufferOutputLogger) -> float:
        """ Returns grading info as a dict """
        logger(f"Grading {get_submission_name(submission.path)}")
        precompiled_submission = self._precompile_submission(submission, logger)
        if precompiled_submission is None:
            return 0
        total_testcase_score = 0
        testcase_results = []
        allowed_tests = self.tests[submission.type]
        if not allowed_tests:
            print(f"No testcases suitable for the submission {submission.path.name} found.")
        submission.type.run_additional_testcase_operations_in_student_dir(submission.dir)
        for test in allowed_tests:
            logger(f"Running '{test.name}'")
            testcase_score, message = test.run(precompiled_submission)
            logger(message)
            testcase_results.append((test.name, message))
            total_testcase_score += testcase_score
        raw_student_score = total_testcase_score / (sum(t.weight for t in allowed_tests) or 1)
        normalized_student_score = raw_student_score * self.config.total_score_to_100_ratio
        self.logger.print_testcase_results_to_results_file(
            submission.path, testcase_results, normalized_student_score, logger
        )
        return normalized_student_score


class AutograderPaths:
    __slots__ = (
        "current_dir",
        "temp_dir",
        "results_dir",
        "output_summary",
        "tests_dir",
        "testcases_dir",
        "extra_dir",
        "input_dir",
        "output_dir",
        "stdout_formatters",
        "default_stdout_formatters",
        "config",
        "default_config",
        "required_dirs",
    )
    current_dir: Path
    temp_dir: Path
    results_dir: Path
    output_summary: Path
    tests_dir: Path

    testcases_dir: Path
    extra_dir: Path
    input_dir: Path
    output_dir: Path
    stdout_formatters: Path
    default_stdout_formatters: Path
    config: Path
    default_config: Path

    required_dirs: tuple

    def __init__(self, current_dir):
        self.current_dir = current_dir
        self.temp_dir = current_dir / "temp"
        self.results_dir = current_dir / "results"
        self.output_summary = current_dir / "grader_output.txt"
        self.tests_dir = current_dir / "tests"

        self.testcases_dir = self.tests_dir / "testcases"
        self.extra_dir = self.tests_dir / "extra"
        self.input_dir = self.tests_dir / "input"
        self.output_dir = self.tests_dir / "output"

        self.stdout_formatters = self.tests_dir / "stdout_formatters.py"
        self.config = self.tests_dir / "config.ini"

        autograder_dir = Path(__file__).parent
        self.default_stdout_formatters = autograder_dir / "default_stdout_formatters.py"
        self.default_config = autograder_dir / "default_config.ini"

        self.required_dirs = (self.testcases_dir,)

    def generate_config(self):
        if not self.config.exists():
            if self.default_config.exists():
                shutil.copy(self.default_config, self.config)
            else:
                raise AutograderError(f"Failed to generate config:'{self.default_config}' not found.")


# Grades single submissions. We use it because multiprocessing.pool only accepts top-level callable objects.
class Runner:
    def __init__(self, grader, lock):
        self.grader: Grader = grader
        self.lock = lock

    def __call__(self, submission: Submission):
        return self.grader.run_on_single_submission(submission, self.lock)


@contextmanager
def temporarily_change_dir(new_dir):
    old_dir = Path.cwd()
    os.chdir(new_dir)
    try:
        yield
    finally:
        os.chdir(old_dir)



from .util.testcase_io import TestCaseIO
from .util.test_helper_formatter import get_formatted_test_helper
import enum
import shutil
from abc import ABC, abstractmethod
from io import StringIO
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import sh
from typing_extensions import Protocol

# TODO: I hate these imports. We should only use relative imports because direct imports might indicate architectural problems.
from autograder.util import get_stderr

from .util.exit_codes import ExitCodeEventType, USED_EXIT_CODES, SYSTEM_RESERVED_EXIT_CODES
from .util.testcase_result_validator import generate_validating_string, validate_output

TEST_HELPERS_DIR = Path(__file__).resolve().parent / "test_helpers"

EMPTY_TESTCASE_IO = TestCaseIO.get_empty_io()


class ShCommand(Protocol):
    """ We use this to imitate sh.Command by ducktyping it """

    def __call__(self, *args: str, **kwargs: Any) -> Optional[sh.RunningCommand]:
        raise NotImplementedError()


def Command(command: str, *args: Any, **kwargs: Any) -> Optional[sh.Command]:
    """ An API for commands that do not throw errors on creation """
    return None if shutil.which(command) is None else sh.Command(command, *args, **kwargs)


class ArgList(enum.Enum):
    SUBMISSION_PRECOMPILATION = "SUBMISSION_PRECOMPILATION_ARGS"
    TESTCASE_PRECOMPILATION = "TESTCASE_PRECOMPILATION_ARGS"
    TESTCASE_COMPILATION = "TESTCASE_COMPILATION_ARGS"


class TestCase(ABC):
    source_suffix = ".source_suffix"  # dummy value
    executable_suffix = ".executable_suffix"  # dummy value
    path: Path
    weight: float

    @property
    @abstractmethod
    def helper_module_name(self) -> str:
        pass

    @classmethod
    @abstractmethod
    def is_installed(cls) -> bool:
        """ Returns True if software necessary to run the testcase is installed on the system """

    @abstractmethod
    def compile_testcase(self, precompiled_submission: Path) -> ShCommand:
        """Compiles student submission and testcase into a single executable
        (or simply returns the command to run the testcase if no further compilation is necessary)

        pwd = temp/student_dir
        """

    def __init__(
        self,
        path: Path,
        timeout: float,
        argument_lists: Dict[ArgList, List[str]],
        anti_cheat_enabled: bool,
        weight: float,
        io: Dict[str, TestCaseIO],
    ):
        self.path = path
        self.timeout = timeout
        self.argument_lists = argument_lists
        self.anti_cheat_enabled = anti_cheat_enabled
        self.weight = weight
        self.max_score = int(weight * 100)

        # Only really works if test name is in snake_case
        self.name = path.stem.replace("_", " ").capitalize()

        self.io = io.get(self.path.stem, EMPTY_TESTCASE_IO)
        self.validating_string = generate_validating_string()

        self.prepend_test_helper()
        # with self.path.open() as f:
        #     print(f.read())
        if anti_cheat_enabled:
            self.precompile_testcase()

        # This is done to hide the contents of testcases and exit codes to the student
        if anti_cheat_enabled:
            with self.path.open("rb") as f:
                self.source_contents = f.read()
            self.path.unlink()

    @classmethod
    def precompile_submission(
        cls,
        submission: Path,
        student_dir: Path,
        possible_source_file_stems: List[str],
        source_is_case_insensitive: bool,
        arglist: List[str],
    ) -> Path:
        """Copies student submission into student_dir and either precompiles
        it and returns the path to the precompiled submission or to the
        copied submission if no precompilation is necesessary

        pwd = temp/student_dir
        """
        destination = (student_dir / possible_source_file_stems[0]).with_suffix(cls.source_suffix)
        shutil.copy(str(submission), str(destination))
        return destination

    def precompile_testcase(self):
        """Replaces the original testcase file with its compiled version,
        thus making reading its contents as plaintext harder.
        Useful in preventing cheating.

        pwd = AutograderPaths.current_dir (i.e. the directory with all submissions)
        """

    @classmethod
    def run_additional_testcase_operations_in_student_dir(cls, student_dir: Path):
        pass

    @classmethod
    def is_a_type_of(cls, file: Path):
        return file.suffix == cls.source_suffix

    def get_path_to_helper_module(self):
        return TEST_HELPERS_DIR / self.helper_module_name

    def run(self, precompiled_submission: Path) -> Tuple[float, str]:
        """ Returns student score and message to be displayed """
        result, message = self._weightless_run(precompiled_submission)

        self.delete_executable_files(precompiled_submission)
        return result * self.weight, message

    def make_executable_path(self, submission: Path) -> Path:
        """ By combining test name and student name, it makes a unique path """
        return submission.with_name(self.path.stem + submission.stem + self.executable_suffix)

    def prepend_test_helper(self):
        """Prepends all of the associated test_helper code to test code

        pwd = AutograderPaths.current_dir (i.e. the directory with all submissions)
        """
        with self.path.open() as f:
            content = f.read()
            final_content = self.get_formatted_test_helper() + "\n" + content
        with self.path.open("w") as f:
            f.write(final_content)

    def get_formatted_test_helper(self) -> str:
        return get_formatted_test_helper(self.get_path_to_helper_module())

    def delete_executable_files(self, precompiled_submission: Path):
        path = self.make_executable_path(precompiled_submission)
        if path.exists():
            path.unlink()

    def delete_source_file(self, source_path: Path):
        if source_path.exists():
            source_path.unlink()

    def _weightless_run(self, precompiled_submission: Path) -> Tuple[float, str]:
        """ Returns student score (without applying testcase weight) and message to be displayed """
        testcase_path = precompiled_submission.with_name(self.path.name)
        if self.anti_cheat_enabled:
            with testcase_path.open("wb") as f:
                f.write(self.source_contents)
        else:
            shutil.copy(self.path, testcase_path)

        try:
            test_executable = self.compile_testcase(precompiled_submission)

        except sh.ErrorReturnCode as e:
            return 0, get_stderr(e, "Failed to compile")
        self.delete_source_file(testcase_path)

        with StringIO() as runtime_output, self.io.input() as runtime_input:
            try:
                result = test_executable(
                    _in=runtime_input,
                    _out=runtime_output,
                    _timeout=self.timeout,
                    _ok_code=USED_EXIT_CODES,
                    _env={"VALIDATING_STRING": self.validating_string},
                )
                if result is None:
                    raise NotImplementedError()
                exit_code = result.exit_code
            except sh.TimeoutException:
                return 0, "Exceeded Time Limit"
            except sh.ErrorReturnCode as e:
                return 0, f"Crashed due to signal {e.exit_code}:\n{e.stderr.decode('UTF-8', 'replace')}\n"
            raw_output = runtime_output.getvalue()
            output, score, output_is_valid = validate_output(raw_output, self.validating_string)
            if not output_is_valid:
                # This  means that either the student used built-in exit function himself
                # or some testcase helper is broken, or a testcase exits itself without
                # the use of helper functions.
                return (
                    0,
                    "None of the helper functions have been called.\n"
                    f"Instead, exit() has been called with exit_code {exit_code}.\n"
                    "It could indicate student cheating or testcases being written incorrectly.",
                )
            elif exit_code == ExitCodeEventType.CHECK_STDOUT:
                if self.io.expected_output_equals(output):
                    return 100, f"{int(100 * self.weight)}/{self.max_score}"
                else:
                    return 0, f"0/{self.max_score} (Wrong output)"
            elif exit_code == ExitCodeEventType.RESULT:
                message = f"{round(score * self.weight, 2)}/{self.max_score}"
                if score == 0:
                    message += " (Wrong answer)"
                return score, message
            elif exit_code in SYSTEM_RESERVED_EXIT_CODES or exit_code < 0:
                # We should already handle this case in try, except block. Maybe we need more info in the error?
                raise NotImplementedError("System error has not been handled.")
            else:
                raise ValueError(f"Unknown system code {exit_code} has not been handled.")


def submission_is_allowed(file: Path, possible_source_file_stems: List[str], source_is_case_insensitive: bool):
    return find_appropriate_source_file_stem(file, possible_source_file_stems, source_is_case_insensitive) is not None


def find_appropriate_source_file_stem(
    file: Path, possible_source_file_stems: List[str], source_is_case_insensitive: bool
) -> Optional[str]:
    file_stem = file.stem
    if source_is_case_insensitive:
        file_stem = file_stem.lower()
    for s in possible_source_file_stems:
        if source_is_case_insensitive:
            if s.lower() in file_stem:
                return s
        else:
            if s in file_stem:
                return s



import configparser
from pathlib import Path
from typing import Any, Callable, Dict, List, Type, TypeVar

from .testcases import ALLOWED_LANGUAGES, ArgList, TestCase, get_testcase_type
from .util import AutograderError

DEFAULT_FILE_STEM = "Homework"


class GradingConfig:
    def __init__(self, testcases_dir: Path, config: Path, default_config: Path):
        self.testcases_dir = testcases_dir
        self._configure_grading(config, default_config)

    def _configure_grading(self, config_path: Path, default_config_path: Path):
        cfg = self._read_config(config_path, default_config_path)

        self.timeouts = parse_config_list(cfg["TIMEOUT"], float)
        self.generate_results = cfg.getboolean("GENERATE_RESULTS")
        self.anti_cheat = cfg.getboolean("ANTI_CHEAT")
        self.parallel_grading_enabled = cfg.getboolean("PARALLEL_GRADING_ENABLED")
        self.multifile_submissions_enabled = cfg.getboolean("MULTIFILE_SUBMISSIONS_ENABLED")

        self.total_points_possible = cfg.getint("TOTAL_POINTS_POSSIBLE")
        self.total_score_to_100_ratio = self.total_points_possible / 100

        language = cfg["PROGRAMMING_LANGUAGE"]
        if language == "AUTO":
            self.testcase_types = self._figure_out_testcase_types()
        else:
            testcase_type = ALLOWED_LANGUAGES[language.lower().strip()]
            self.testcase_types = {testcase_type.source_suffix: testcase_type}

        self.assignment_name = cfg["ASSIGNMENT_NAME"]

        source = cfg["POSSIBLE_SOURCE_FILE_STEMS"].strip()
        if source == "AUTO":
            source = DEFAULT_FILE_STEM
            self.auto_source_file_name_enabled = True
        else:
            self.auto_source_file_name_enabled = False
        self.source_file_stem_is_case_insensitive = cfg.getboolean("SOURCE_FILE_NAME_IS_CASE_INSENSITIVE")
        self.possible_source_file_stems = source.replace(" ", "").split(",")

        self.testcase_weights = parse_config_list(cfg["TESTCASE_WEIGHTS"], float)

        # TODO: Name me better. The name is seriously bad
        self.cli_argument_lists = parse_arglists(cfg)

    @staticmethod
    def _read_config(path_to_user_config: Path, path_to_default_config: Path) -> configparser.SectionProxy:
        default_parser = configparser.ConfigParser()
        default_parser.read(path_to_default_config)

        user_parser = configparser.ConfigParser()
        user_parser.read_dict(default_parser)
        user_parser.read(path_to_user_config)

        return user_parser["CONFIG"]

    def generate_arglists(self, file_name: str) -> Dict[ArgList, List[str]]:
        arglist = {}
        for arglist_index, arglists_per_file in self.cli_argument_lists.items():
            if file_name in arglists_per_file:
                arglist[arglist_index] = arglists_per_file[file_name]
            elif "ALL" in arglists_per_file:
                arglist[arglist_index] = arglists_per_file["ALL"]
            else:
                arglist[arglist_index] = tuple()
        return arglist

    def _figure_out_testcase_types(self) -> Dict[str, Type[TestCase]]:
        testcase_types = {}
        for testcase in self.testcases_dir.iterdir():
            testcase_type = get_testcase_type(testcase)
            if testcase_type is not None:
                testcase_types[testcase_type.source_suffix] = testcase_type

        if testcase_types:
            return testcase_types
        else:
            raise AutograderError(f"Couldn't discover any testcases with supported suffixes in {self.testcases_dir}")


T = TypeVar("T")


def parse_config_list(config_line: str, value_type: Callable[[Any], T]) -> Dict[str, T]:
    """Reads in a config line in the format: 'file_name:value, file_name:value, ALL:value'
    ALL sets the default value for all other entries
    """
    config_entries = {}
    for config_entry in config_line.split(","):
        if config_entry.strip():
            testcase_name, value = config_entry.strip().split(":")
            config_entries[testcase_name.strip()] = value_type(value)
    return config_entries


def parse_arglists(cfg: configparser.SectionProxy) -> Dict[ArgList, Dict[str, List[str]]]:
    argument_lists = {n: {} for n in ArgList}
    convert_to_arglist = lambda s: str(s).replace("  ", " ").strip().split()
    for arg_list_enum in ArgList:
        arg_list = parse_config_list(cfg[arg_list_enum.value], convert_to_arglist)
        for testcase_name, args in arg_list.items():
            argument_lists[arg_list_enum][testcase_name] = args
    return argument_lists



from pathlib import Path
from typing import Dict, Iterable, Optional, Type

from .abstract_base_class import ArgList, TestCase
from .c import CTestCase
from .cpp import CPPTestCase
from .java import JavaTestCase
from .python import PythonTestCase


def _is_installed(language_name: str, testcase: Type[TestCase]) -> bool:
    """ Useful for logging """
    if testcase.is_installed():
        return True
    else:
        print(f"Utilities for running {language_name} are not installed. Disabling it.")
        return False


class Submission:
    __slots__ = "path", "type", "dir"

    def __init__(self, file: Path, testcase_type: Type[TestCase], temp_dir: Path):
        self.path = file
        self.type = testcase_type
        self.dir = temp_dir / file.name
        self.dir.mkdir()


ALLOWED_LANGUAGES: Dict[str, Type[TestCase]] = {
    "c": CTestCase,
    "java": JavaTestCase,
    "python": PythonTestCase,
    "c++": CPPTestCase,
}

ALLOWED_LANGUAGES = {k: v for k, v in ALLOWED_LANGUAGES.items() if _is_installed(k, v)}


def get_testcase_type(
    file: Path, allowed_types: Iterable[Type[TestCase]] = ALLOWED_LANGUAGES.values()
) -> Optional[Type[TestCase]]:
    for testcase_type in allowed_types:
        if testcase_type.is_a_type_of(file):
            return testcase_type


from autograder.util import AutograderError
import re
import shutil
from pathlib import Path
from typing import List

from .abstract_base_class import TEST_HELPERS_DIR, ArgList, Command, TestCase, find_appropriate_source_file_stem

PUBLIC_CLASS_MATCHER = re.compile(r"public(?:\w|\s)+class(?:\w|\s)+({)")
PATH_TO_JNA_FILE = TEST_HELPERS_DIR / "extra" / "jna.jar"
PATH_TO_SECURITY_MANAGER_FILE = TEST_HELPERS_DIR / "extra" / "NoReflectionAndEnvVarsSecurityManager.class"


class JavaTestCase(TestCase):
    """Please, ask students to remove their main as it can theoretically
    generate errors (not sure how though).
    Java doesn't support testcase precompilation because it must always
    link files on compilation.
    """

    source_suffix = ".java"
    executable_suffix = ""
    helper_module_name = "TestHelper.java"
    compiler = Command("javac")
    virtual_machine = Command("java")

    @classmethod
    def is_installed(cls) -> bool:
        return cls.compiler is not None and cls.virtual_machine is not None

    @classmethod
    def precompile_submission(
        cls,
        submission: Path,
        student_dir: Path,
        possible_source_file_stems: List[str],
        source_is_case_insensitive: bool,
        arglist,
    ):
        stem = find_appropriate_source_file_stem(submission, possible_source_file_stems, source_is_case_insensitive)
        if stem is None:
            raise AutograderError(
                f"Submission {submission} has an inappropriate file name. Please, specify POSSIBLE_SOURCE_FILE_STEMS in config.ini"
            )
        copied_submission = super().precompile_submission(
            submission, student_dir, [stem], source_is_case_insensitive, arglist
        )
        try:
            cls.compiler(copied_submission, *arglist)
        finally:
            copied_submission.unlink()

        return copied_submission.with_suffix(".class")

    def compile_testcase(self, precompiled_submission: Path):
        new_self_path = precompiled_submission.with_name(self.path.name)
        self.compiler(
            new_self_path,
            "-cp",
            f".:*",
            *self.argument_lists[ArgList.TESTCASE_COMPILATION],
        )
        return lambda *args, **kwargs: self.virtual_machine("-cp", ".:*", self.path.stem, *args, **kwargs)

    @classmethod
    def run_additional_testcase_operations_in_student_dir(cls, student_dir: Path):
        shutil.copyfile(PATH_TO_JNA_FILE, student_dir / PATH_TO_JNA_FILE.name)
        shutil.copyfile(PATH_TO_SECURITY_MANAGER_FILE, student_dir / PATH_TO_SECURITY_MANAGER_FILE.name)

    def delete_executable_files(self, precompiled_submission: Path):
        for p in precompiled_submission.parent.iterdir():
            if p.suffix == ".class" and self.path.stem in p.stem:
                p.unlink()

    def prepend_test_helper(self):
        """Puts private TestHelper at the end of testcase class.
        This is quite a crude way to do it but it is the easiest
        I found so far.
        """
        with open(self.path) as f:
            content = f.read()
        final_content = self._add_at_the_beginning_of_public_class(self.get_formatted_test_helper(), content)
        final_content = (
            f"""import com.sun.jna.Library;
                import com.sun.jna.Native;
                import java.lang.reflect.Field;
                import java.util.Map;
                import java.util.HashMap;
            """
            + final_content
        )
        with open(self.path, "w") as f:
            f.write(final_content)

    def _add_at_the_beginning_of_public_class(self, helper_class: str, java_file: str):
        """ Looks for the first bracket of the first public class and inserts test helper next to it """
        # This way is rather crude and can be prone to errors if left untested,
        # but java does not really leave us any other way to do it.
        match = PUBLIC_CLASS_MATCHER.search(java_file)
        if match is None:
            raise ValueError(f"Public class not found in {self.path}")
        else:
            main_class_index = match.end(1)

        return "".join(
            [
                java_file[:main_class_index],
                "\n" + helper_class + "\n",
                java_file[main_class_index:],
            ]
        )



from collections import namedtuple
from autograder.testcases.util.exit_codes import ExitCodeEventType
from pathlib import Path
import shutil
from typing import List

import sh
from .abstract_base_class import Command, TestCase, submission_is_allowed

POSSIBLE_MAKEFILE_NAMES = "GNUmakefile", "makefile", "Makefile"
DUMMY_SH_COMMAND_RESULT_CLASS = namedtuple("ShCommandResult", "exit_code")


def is_multifile_submission(submission_dir: Path, possible_source_file_stems, source_is_case_insensitive):
    contents = list(submission_dir.iterdir())
    while len(contents) == 1 and contents[0].is_dir():
        contents = list(contents[0].iterdir())

    for f in submission_dir.iterdir():
        if _is_makefile(f) or submission_is_allowed(f, possible_source_file_stems, source_is_case_insensitive):
            return True
    else:
        return False


def _is_makefile(path: Path):
    return not path.is_dir() and path.name in POSSIBLE_MAKEFILE_NAMES


class MultifileTestCase(TestCase):
    helper_module_name = ""
    compiler = Command("make")

    @classmethod
    def is_installed(cls) -> bool:
        """We assume all utilities necessary are installed because we can't possibly check for all of them."""
        return cls.compiler is not None

    @classmethod
    def precompile_submission(
        cls,
        submission: Path,
        student_dir: Path,
        possible_source_file_stems: str,
        source_is_case_insensitive: bool,
        arglist,
    ):
        """ pwd = temp/student_dir """
        _copy_submission_contents_into_student_dir(submission, student_dir)
        try:
            cls.compiler()
        except sh.ErrorReturnCode as e:
            if "no makefile found" in str(e):
                pass
            else:
                raise e
        return _find_submission_executable(student_dir, possible_source_file_stems, source_is_case_insensitive)

    def compile_testcase(self, precompiled_submission: Path):
        return lambda *a, **kw: _run_multifile_testcase(precompiled_submission, *a, **kw)

    def prepend_test_helper(self):
        """ We don't need TestHelper when we are only checking inputs/outputs """

    def prepend_test_helper(self):
        """We don't need test helper because we don't have a testcase file"""
        pass

    def delete_executable_files(self, precompiled_submission: Path):
        """ We don't compile each testcase separately so we don't need to clean up executable files """
        pass

    def delete_source_file(self, source_path: Path):
        """ There is no testcase file """
        pass


def _run_multifile_testcase(precompiled_submission: Path, *args, **kwargs):
    sh.Command(precompiled_submission)(*args, **kwargs)
    return DUMMY_SH_COMMAND_RESULT_CLASS(ExitCodeEventType.CHECK_STDOUT)


def _copy_submission_contents_into_student_dir(submission, student_dir):
    contents: List[Path] = list(submission.iterdir())
    while len(contents) == 1 and contents[0].is_dir():
        contents = list(contents[0].iterdir())
    for f in contents:
        new_path = student_dir / f.name
        if f.is_dir():
            shutil.copytree(f, new_path)
        else:
            shutil.copyfile(f, new_path)


def _find_submission_executable(student_dir: Path, possible_source_file_stems, source_is_case_insensitive):
    for f in student_dir.iterdir():
        if submission_is_allowed(f, possible_source_file_stems, source_is_case_insensitive):
            return f
